#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>

#include <oslib/osfile.h>
#include <oslib/osfscontrol.h>
#include <oslib/osgbpb.h>
#include <oslib/jpeg.h>
#include <oslib/quit.h>
#include <oslib/stringset.h>
#include <oslib/writablefield.h>
#include <oslib/window.h>
#include <oslib/colourtrans.h>
#include <oslib/iconbar.h>
#include <oslib/hourglass.h>
#include <oslib/saveas.h>
#include <oslib/hal.h>
#include <Systypes.h>
#include <Event.h>

#include <unixlib/local.h>

static int toolbox_quit;
static messagetrans_control_block messages_file;
static toolbox_o main_window;
static wimp_w w;
static jpeg_image *image = NULL;
static int width, height, size;
static int converting;
static char imagefile[256];

toolbox_action_list toolbox_events;     // holds toolbox event codes
wimp_message_list wimp_messages;

static void toolbox_poll(int active);

typedef struct toolbox_message_list {
	int first;
	int rest[];
} toolbox_message_list;

typedef union message_list {
	toolbox_message_list toolbox;
	wimp_message_list wimp;
	toolbox_action_list Toolbox;
} message_list;

typedef enum toolbox_result {
	TOOLBOX_RESULT_OK,
	TOOLBOX_RESULT_NOMEM,
	TOOLBOX_RESULT_OPEN,
	TOOLBOX_RESULT_WRITE
} toolbox_result;

void *minst;

#ifndef UNUSED
#define UNUSED(x) ((x)=(x))
#endif

#define PTR_WIMP_MESSAGE_LIST(l) ((wimp_message_list *) (message_list *) (l))
#define PTR_TOOLBOX_ACTION_LIST(l) ((toolbox_action_list *) (message_list *) (l))


/* Wimp Messages we're interested in */
/* static toolbox_message_list messages = {
	message_DATA_LOAD,
	{
		message_QUIT,
		0
	}
}; */

/* Toolbox events we're interested in */
/* static toolbox_message_list tbcodes = {
	Convert_Font,
	{
		action_ERROR,
		action_QUIT_QUIT,
		0
	}
};  */

static void toolbox_init(int argc, char **argv);
static void toolbox_exit(void);
static void register_toolbox_handlers(void);
static osbool toolbox_event_quit(bits event_code, toolbox_action *event,toolbox_block *id_block, void *handle);
static osbool redraw_window(wimp_event_no event_code,wimp_block *block, toolbox_block *id_block,void *handle);
static osbool toolbox_event_error(bits event_code, toolbox_action *event,
		toolbox_block *id_block, void *handle);
static void register_message_handlers(void);
static int wimp_message_quit(wimp_message *message,void *handle);
static int wimp_message_dataload(wimp_message *message,void *handle);
static void bitmap_load(char *filename);
static int bitmap_convert(char *filename);
static osbool close_window(wimp_event_no event_code, wimp_block *block, toolbox_block *id_block, void *handle);
static void destroy_window(void);
extern double get_timerValue(void);
osbool toolbox_saveas(bits event_code, toolbox_action *event, toolbox_block *id_block, void *handle);

int main(int argc, char **argv)
{
	toolbox_init(argc, argv);

	while (!toolbox_quit)
		toolbox_poll(0);

	toolbox_exit();

	return 0;
}


void toolbox_init(int argc, char **argv)
{
  os_error *error;

	toolbox_block toolbox_block;

    wimp_messages.messages[0]=0;
    toolbox_events.action_nos[0]=0;

	UNUSED(argc);
	UNUSED(argv);

	event_initialise(&toolbox_block);

	event_set_mask(wimp_MASK_NULL |
			wimp_MASK_LEAVING |
			wimp_MASK_ENTERING |
			wimp_QUEUE_MOUSE |
			wimp_MASK_LOSE |
			wimp_MASK_GAIN);

	register_toolbox_handlers();
	register_message_handlers();

	error = xtoolbox_initialise(0, 310,
				&wimp_messages,
				&toolbox_events,
				"<JPEGView$Dir>", &messages_file,
				&toolbox_block, NULL, NULL, NULL);
	if (error) {
		fprintf(stderr, "toolbox_initialise: 0x%x: %s\n",
			error->errnum, error->errmess);
			exit(1);
	}


	error = xtoolbox_create_object(0, (toolbox_id) "Window", &main_window);
	if (error) {
		fprintf(stderr, "toolbox_create_object: 0x%x: %s\n",
			error->errnum, error->errmess);
		 exit(1);


	  }

}

static void bitmap_load(char *filename)
{
  os_error *error;
  BBox extent;
  osbool success;
  double startTime;
  double endTime;


	 if (main_window)
	   destroy_window();

	   xhourglass_on();
	   startTime = get_timerValue();

	error = xtoolbox_create_object(0, (toolbox_id) "Window", &main_window);
	if (error) {
		fprintf(stderr, "toolbox_create_object: 0x%x: %s\n",
			error->errnum, error->errmess);
		 exit(1);


	  }
	  	error = xwindow_set_title(0, main_window, filename);
	if (error != NULL) {
		fprintf(stderr, "displayfield_set_value: 0x%x: %s\n",
			error->errnum, error->errmess);
	}

	success = event_register_wimp_handler(event_ANY,
			wimp_REDRAW_WINDOW_REQUEST, redraw_window, NULL);
	if (success == FALSE)
	fprintf(stderr, "registering RedrawWindow failed\n");

	success = event_register_wimp_handler(event_ANY,
			wimp_CLOSE_WINDOW_REQUEST, close_window, NULL);
	if (success == FALSE)
	fprintf(stderr, "registering CloseWindow failed\n");

	if (bitmap_convert(filename) == 0)
	{
	error = xosfile_read_no_path(filename, 0, 0, 0, &size, 0);
 	if (error) {
		fprintf(stderr, "osfile_read_path: 0x%x: %s\n",
			error->errnum, error->errmess);
			exit(1);
	}

	fprintf(stderr, "FileSize: %d\n", size);

 /* Use calloc instead of malloc so the jpeg is padded on a word boundary */

 image = (jpeg_image *)calloc(1, size);

 error = xosfile_load_stamped_no_path(filename, (byte *) image, 0, 0, 0, 0, 0);
	if (error) {
		fprintf(stderr, "xosfile_load_stamped_no_path: 0x%x: %s\n",
		error->errnum, error->errmess);
		if (image) {
		free(image);
		image = NULL;
		}
		xhourglass_off();
		 exit(1);
}

  error = xjpeginfo_dimensions((jpeg_image const *) image, 0, 0, &width, &height, 0, 0, 0);
	if (error) {
		fprintf(stderr, "xjpeginfo_dimensions: 0x%x: %s\n",
			error->errnum, error->errmess);
		 exit(1);
}


xwindow_get_extent(0, main_window, &extent);
extent.y0 = 0 - (height * 2);
extent.x1 = width * 2;
xwindow_set_extent(0, main_window, &extent);
	fprintf(stderr, "x0: %d y0: %d x1: %d y1: %d\n", extent.x0, extent.y0, extent.x1, extent.y1);
 xtoolbox_show_object(0, main_window, toolbox_POSITION_DEFAULT, 0, NULL, (toolbox_c)NULL);

}

	 endTime = get_timerValue();
	       fprintf(stderr, "Process time total %fs for %s\n", endTime - startTime, filename);

xhourglass_off();

}
static void toolbox_poll(int active)
{
	int event;
	unsigned int mask;
	wimp_block block;

	if (active || converting) {
		event_set_mask(0x3972);
		event_poll(&event, &block, 0);
	} else {
		event_get_mask(&mask);
		event_set_mask(mask | wimp_MASK_NULL);
		event_poll(&event, &block, 0);
	}
}

void toolbox_exit(void)
{
	if (image) {
	  free(image);
	  image = NULL;
	  }

}

/**
 * Register event handlers with the toolbox
 */
void register_toolbox_handlers(void)
{
	 osbool success;


 	success = event_register_toolbox_handler(event_ANY,
			action_QUIT_QUIT, toolbox_event_quit, NULL);
	if (success == FALSE)
		fprintf(stderr, "registering quit_QUIT failed\n");

	success = event_register_toolbox_handler(event_ANY,
			action_ERROR, toolbox_event_error, NULL);
	if (success == FALSE)
	fprintf(stderr, "registering action_ERROR failed\n");

	success = event_register_toolbox_handler(event_ANY,
			action_SAVE_AS_ABOUT_TO_BE_SHOWN, toolbox_saveas, NULL);
	if (success == FALSE)
		fprintf(stderr, "registering action_SAVE_AS_ABOUT_TO_BE_SHOWN failed\n");

}
static osbool close_window(wimp_event_no event_code, wimp_block *block, toolbox_block *id_block, void *handle)
{
  destroy_window();
  return TRUE;

}

static void destroy_window(void)
{
  event_deregister_wimp_handler(event_ANY,
			wimp_REDRAW_WINDOW_REQUEST, redraw_window, NULL);
  event_deregister_wimp_handler(event_ANY,
			wimp_CLOSE_WINDOW_REQUEST, close_window, NULL);
  toolbox_delete_object(0, main_window);

  main_window = NULL;

  if (image) {

    free(image);
    image = NULL;
    }

}


osbool redraw_window(wimp_event_no event_code, wimp_block *block, toolbox_block *id_block, void *handle)
{
 	UNUSED(event_code);
//	UNUSED(event);
//	UNUSED(block);
	UNUSED(id_block);
	UNUSED(handle);
  osbool more;
  wimp_draw redraw;
  int x0, y0;
  BBox extent;
   xwindow_get_wimp_handle(0, main_window, &w);
   redraw.w = w;
   xwimp_redraw_window(&redraw, &more);
	os_factors factors;
  x0 = redraw.box.xmin - redraw.xscroll;
  y0 = redraw.box.ymax - redraw.yscroll;
  xwindow_get_extent(0, id_block->this_obj, &extent);
    y0 += extent.ymin;

        factors.xmul=width*2;
        factors.ymul=height*2;
        factors.xdiv=width*2;
        factors.ydiv=height*2;
        while (more)
        {
	         xjpeg_plot_scaled(image, x0, y0, &factors, size, 0);
	         xwimp_get_rectangle(&redraw, &more);
	}


 return TRUE;
}

#define HANDLE_FATAL_ERROR(ret)\
    if(ret)\
        {\
            fprintf(stderr,"%s: libraw  %s\n",filename,libraw_strerror(ret));\
            if(LIBRAW_FATAL_ERROR(ret))\
              libraw_recycle(iprc); \
              libraw_close(iprc); \
                return 1;               \
        }\

#define HANDLE_ALL_ERRORS(ret)\
    if(ret)\
        {\
            fprintf(stderr,"%s: libraw  %s\n",filename,libraw_strerror(ret));\
              libraw_recycle(iprc); \
              libraw_close(iprc); \
            return 1;                                                   \
        }\


static int bitmap_convert(char *filename)
{

    return 0;

}

/**
 * Handle quit events
 */
osbool toolbox_event_quit(bits event_code, toolbox_action *event,
		toolbox_block *id_block, void *handle)
{
	UNUSED(event_code);
	UNUSED(event);
	UNUSED(id_block);
	UNUSED(handle);

	toolbox_quit = 1;

	return TRUE;
}

/**
 * Handle toolbox errors
 */
osbool toolbox_event_error(bits event_code, toolbox_action *event,
		toolbox_block *id_block, void *handle)
{
	toolbox_action_error_block *error =
			(toolbox_action_error_block *) event;

	UNUSED(event_code);
	UNUSED(id_block);
	UNUSED(handle);

	fprintf(stderr, "toolbox error: 0x%x: %s\n",
		error->errnum, error->errmess);

	return TRUE;
}

/**
 * Register message handlers
 */
void register_message_handlers(void)
{
	osbool success;

	success = event_register_message_handler(message_QUIT,
			wimp_message_quit, NULL);
	if (success == FALSE)
		fprintf(stderr, "registering message_QUIT handler failed\n");

	success = event_register_message_handler(message_DATA_LOAD,
			wimp_message_dataload, NULL);
	if (success == FALSE)
		fprintf(stderr,
		"registering message_DATA_LOAD handler failed\n");

        success = event_register_message_handler(message_DATA_OPEN,
			wimp_message_dataload, NULL);
	if (success == FALSE)
		fprintf(stderr,
			"registering message_DATA_OPEN handler failed\n");

}

/**
 * Handle quit messages
 */
osbool wimp_message_quit(wimp_message *message, void *handle)
{
	UNUSED(message);
	UNUSED(handle);

	toolbox_quit = 1;

	return TRUE;
}

/**
 * Handle dataload messages
 */
osbool wimp_message_dataload(wimp_message *message, void *handle)
{
	os_error *error;
	wimp_message_data_xfer *dl = &message->data.data_xfer;

	UNUSED(handle);

  if (dl->file_type > 0xfff)
    return FALSE;

    if (dl->w != wimp_ICON_BAR)
    {
    switch(dl->file_type)
    {
    case 0xc85 : // Jpeg
     break;
    default:
      return FALSE;
    }
    }

	message->action = message_DATA_LOAD_ACK;
	message->your_ref = message->my_ref;

	error = xwimp_send_message(wimp_USER_MESSAGE, message, message->sender);
	if (error != NULL) {
		fprintf(stderr, "wimp_send_message: 0x%x: %s\n",
			error->errnum, error->errmess);
	  }

	strcpy(imagefile, dl->file_name);
	bitmap_load(dl->file_name);

	return TRUE;
}
/**
 * Handle saveas events
 */
osbool toolbox_saveas(bits event_code, toolbox_action *event,
		toolbox_block *id_block, void *handle)
  {

    os_error *error;

	UNUSED(event_code);
	UNUSED(event);
//	UNUSED(id_block);
	UNUSED(handle);

          error = xsaveas_set_file_size(0, id_block->this_obj, size - sizeof(int));
    	  if (error) {
		fprintf(stderr, "xsaveas_set_file_size: 0x%x: %s\n",
			error->errnum, error->errmess);
	  }

          error = xsaveas_set_data_address(0, id_block->this_obj, (byte *) image, size,
          				(byte *) image, size);
       	  if (error) {
		fprintf(stderr, "xsaveas_set_data_address: 0x%x: %s\n",
			error->errnum, error->errmess);
	  }

	return TRUE;
  }
